#include <blkid/blkid.h>
#include <errno.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mount.h>
#include <unistd.h>

#define UNSAFE_C_START \
  _Pragma ("GCC diagnostic push") _Pragma ("GCC diagnostic ignored \"-Wsign-conversion\"")

#define UNSAFE_C_END _Pragma ("GCC diagnostic pop")

#define pure_ __attribute__ ((pure))
#define autofree __attribute__ ((cleanup (cleanup_free)))
#define autofclose __attribute__ ((cleanup (cleanup_fclose)))

static bool enable_debug = false;

__attribute__ ((__format__ (printf, 1, 2))) static void
debug (const char *format, ...)
{
  if (enable_debug)
    {
      va_list args;
      va_start (args, format);
      vprintf (format, args);
      va_end (args);
    }
}

__attribute__ ((__format__ (printf, 1, 2))) static void
perrorf (const char *format, ...)
{
  va_list args;
  va_start (args, format);
  vfprintf (stderr, format, args);
  fprintf (stderr, ": %s\n", strerror (errno));
  va_end (args);
}

static inline void
cleanup_free (void *p)
{
  free (*(void **)p);
}

static inline void
cleanup_fclose (FILE **stream)
{
  if (*stream)
    fclose (*stream);
}

static char *
read_proc_cmdline (void)
{
  autofclose FILE *f = fopen ("/proc/cmdline", "r");
  if (!f)
    {
      perror ("Failed to open /proc/cmdline");
      return NULL;
    }

  char *cmdline = NULL;
  size_t len;

  /* Note that /proc/cmdline will not end in a newline, so getline
   * will fail unless we provide a length.
   */
  if (getline (&cmdline, &len, f) < 0)
    {
      perror ("Failed to read /proc/cmdline");
      return NULL;
    }

  /* ... but the length will be the size of the malloc buffer, not
   * strlen().  Fix that.
   */
  len = strlen (cmdline);
  if (cmdline[len - 1] == '\n')
    cmdline[len - 1] = '\0';

  return cmdline;
}

static char *
find_proc_cmdline_key (const char *cmdline, const char *key)
{
  const size_t key_len = strlen (key);
  for (const char *iter = cmdline; iter;)
    {
      const char *next = strchr (iter, ' ');
      if (strncmp (iter, key, key_len) == 0 && iter[key_len] == '=')
        {
          const char *start = iter + key_len + 1;
          if (next)
            {
              UNSAFE_C_START
              const size_t len = next - start;
              UNSAFE_C_END
              return strndup (start, len);
            }

          return strdup (start);
        }

      if (next)
        next += strspn (next, " ");

      iter = next;
    }

  return NULL;
}

static bool pure_
has_proc_cmdline_flag (const char *cmdline, const char *key)
{
  const size_t key_len = strlen (key);
  for (const char *iter = cmdline; iter;)
    {
      const char *next = strchr (iter, ' ');
      size_t val_len = 0;

      if (next)
        {
          UNSAFE_C_START
          const size_t len = next - iter;
          UNSAFE_C_END
          val_len = len;
        }
      else
        val_len = strlen (iter);

      if (val_len == key_len && strncmp (iter, key, key_len) == 0)
        return true;

      if (next)
        next += strspn (next, " ");

      iter = next;
    }

  return false;
}

static bool
exists (const char *const fname)
{
  return access (fname, F_OK) == 0;
}

static int
attempt_mount (const char *const device, const unsigned long sysroot_mount_flags,
               const char *const fstype)
{
  debug ("Mounting /sysroot from %s (%s)\n", device,
         ((sysroot_mount_flags & MS_RDONLY) != 0) ? "readonly" : "readwrite");
  if (mount (device, "/sysroot", fstype, sysroot_mount_flags, NULL) == -1)
    {
      perrorf ("mount(\"%s\", \"/sysroot\", \"%s\", %lu, NULL)", device, fstype,
               sysroot_mount_flags);
      if (!exists (device))
        perrorf ("%s doesn't exist", device);

      if (!exists ("/sysroot"))
        perrorf ("/sysroot doesn't exist");

      return EXIT_FAILURE;
    }

  debug ("Mount /sysroot success\n");

  return EXIT_SUCCESS;
}

int
main (void)
{
  const autofree char *cmdline = read_proc_cmdline ();
  if (cmdline == NULL)
    return EXIT_FAILURE;

  enable_debug = has_proc_cmdline_flag (cmdline, "mount-sysroot.debug");
  autofree char *device = find_proc_cmdline_key (cmdline, "mount-sysroot.root");
  if (device == NULL)
    device = find_proc_cmdline_key (cmdline, "root");

  if (device == NULL)
    {
      perror ("No mount-sysroot.root= or root= karg present in /proc/cmdline");
      return EXIT_FAILURE;
    }

  autofree char *fstype = find_proc_cmdline_key (cmdline, "mount-sysroot.rootfstype");
  if (fstype == NULL)
    fstype = find_proc_cmdline_key (fstype, "rootfstype");

  unsigned long sysroot_mount_flags = 0;
  if (has_proc_cmdline_flag (cmdline, "mount-sysroot.ro")
      || (!has_proc_cmdline_flag (cmdline, "mount-sysroot.rw")
          && has_proc_cmdline_flag (cmdline, "ro")))
    sysroot_mount_flags |= MS_RDONLY;

  if (strncmp (device, "PARTUUID=", 9) == 0 || strncmp (device, "PARTLABEL=", 10) == 0)
    {
      device = blkid_get_devname (NULL, device, NULL);
      if (device == NULL)
        {
          perror ("Error getting device name");
          return EXIT_FAILURE;
        }
    }

  return attempt_mount (device, sysroot_mount_flags, fstype);
}
